"""Module for the discrete transfer entropy estimator."""

from abc import ABC

from numpy import ndarray

from ..base import (
    ConditionalTransferEntropyEstimator,
    TransferEntropyEstimator,
)
from ..mixins import DiscreteTEMixin
from ..utils.discrete_transfer_entropy import combined_te_form
from ..utils.ordinal import reduce_joint_space
from ..utils.te_slicing import cte_observations, te_observations
from ... import Config
from ...utils.types import LogBaseType


class BaseDiscreteTEEstimator(DiscreteTEMixin, ABC):
    """Base class for discrete transfer entropy estimators.

    Attributes
    ----------
    source, dest : array-like
        The source (X) and destination (Y) data used to estimate the transfer entropy.
    cond : array-like or tuple of array-like, optional
        The conditional data used to estimate the conditional transfer entropy.
        Multiple RVs can be passed as tuples and will be treated as joint distributions.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other (multiple of
        ``step_size``).
        Delay/lag/shift between the variables, representing propagation time.
        Assumed time taken by info to transfer from source to destination.
        Not compatible with the ``cond`` parameter / conditional TE.
        Alternatively called `offset`.
    step_size : int, optional
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len : int, optional
        Number of past observations to consider for the source and destination data.
    cond_hist_len : int, optional
        Number of past observations to consider for the conditional data.
        Only used for conditional transfer entropy.
    """

    def __init__(
        self,
        source,
        dest,
        *,  # Enforce keyword-only arguments
        cond=None,
        prop_time: int = 0,
        step_size: int = 1,
        src_hist_len: int = 1,
        dest_hist_len: int = 1,
        cond_hist_len: int = 1,
        offset: int = None,
        base: LogBaseType = Config.get("base"),
        **kwargs,
    ):
        """Initialize the BaseDiscreteTEEstimator.

        Parameters
        ----------
        source, dest : array-like
            The source (X) and destination (Y) data used to estimate the transfer entropy.
        cond : array-like or tuple of array-like, optional
            The conditional data used to estimate the conditional transfer entropy.
            Multiple RVs can be passed as tuples and will be treated as joint distributions.
        prop_time : int, optional
            Number of positions to shift the data arrays relative to each other (multiple of
            ``step_size``).
            Delay/lag/shift between the variables, representing propagation time.
            Assumed time taken by info to transfer from source to destination
            Not compatible with the ``cond`` parameter / conditional TE.
            Alternatively called `offset`.
        step_size : int, optional
            Step size between elements for the state space reconstruction.
        src_hist_len, dest_hist_len : int, optional
            Number of past observations to consider for the source and destination data.
        cond_hist_len : int, optional
            Number of past observations to consider for the conditional data.
            Only used for conditional transfer entropy.
        """
        self.source = source
        self.dest = dest
        if cond is None:
            super().__init__(
                source,
                dest,
                prop_time=prop_time,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                step_size=step_size,
                offset=offset,
                base=base,
                **kwargs,
            )
        else:
            if isinstance(cond, (tuple, ndarray)):
                cond = reduce_joint_space(cond)
            super().__init__(
                source,
                dest,
                cond=cond,
                step_size=step_size,
                src_hist_len=src_hist_len,
                dest_hist_len=dest_hist_len,
                cond_hist_len=cond_hist_len,
                prop_time=prop_time,
                offset=offset,
                base=base,
                **kwargs,
            )
        self._check_data_te()


class DiscreteTEEstimator(BaseDiscreteTEEstimator, TransferEntropyEstimator):
    """Estimator for discrete transfer entropy.

    Attributes
    ----------
    source, dest : array-like
        The source (X) and destination (Y) data used to estimate the transfer entropy.
    prop_time : int, optional
        Number of positions to shift the data arrays relative to each other (multiple of
        ``step_size``).
        Delay/lag/shift between the variables, representing propagation time.
        Assumed time taken by info to transfer from source to destination.
        Alternatively called `offset`.
    step_size : int
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len : int
        Number of past observations to consider for the source and destination data.
    """

    def _calculate(self):
        """Estimate the Discrete Transfer Entropy."""
        return combined_te_form(
            te_observations,
            self.source,
            self.dest,
            local=False,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            step_size=self.step_size,
            permute_src=self.permute_src,
            resample_src=self.resample_src,
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return combined_te_form(
            te_observations,
            self.source,
            self.dest,
            local=True,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            step_size=self.step_size,
            permute_src=self.permute_src,
            resample_src=self.resample_src,
        )


class DiscreteCTEEstimator(
    BaseDiscreteTEEstimator, ConditionalTransferEntropyEstimator
):
    """Estimator for discrete conditional transfer entropy.

    Attributes
    ----------
    source, dest, cond : array-like
        The source (X), destination (Y), and conditional (Z) data used to estimate the
        conditional transfer entropy.
    step_size : int
        Step size between elements for the state space reconstruction.
    src_hist_len, dest_hist_len, cond_hist_len : int, optional
        Number of past observations to consider for the source, destination,
        and conditional data.
    prop_time : int, optional
        Not compatible with the ``cond`` parameter / conditional TE.
    """

    def _calculate(self):
        """Estimate the Discrete Transfer Entropy."""
        return combined_te_form(
            cte_observations,
            self.source,
            self.dest,
            self.cond,
            local=False,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            cond_hist_len=self.cond_hist_len,
            step_size=self.step_size,
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return combined_te_form(
            cte_observations,
            self.source,
            self.dest,
            self.cond,
            local=True,
            log_func=self._log_base,
            src_hist_len=self.src_hist_len,
            dest_hist_len=self.dest_hist_len,
            cond_hist_len=self.cond_hist_len,
            step_size=self.step_size,
        )



"""
Explicit tests for the Kraskov-Stoegbauer-Grassberger (KSG) transfer entropy estimator.
"""

import numpy as np
import pytest
from numpy import ndarray, inf

from infomeasure import transfer_entropy
from infomeasure.estimators.transfer_entropy import KSGTEEstimator, KSGCTEEstimator
from tests.conftest import (
    generate_autoregressive_series,
    generate_autoregressive_series_condition,
)


@pytest.mark.parametrize(
    "rng_int,k,ksg_id,minkowski_p,base,expected",
    [
        (5, 4, 1, 2, "e", -0.1464972645512768),
        (5, 4, 1, 3, "e", -0.07711205666343),
        (5, 4, 2, 3, "e", -0.32711205666),
        (5, 4, 1, inf, "e", 0.09269348225),
        (5, 16, 1, 2, "e", -0.09411416419535731),
        (5, 16, 1, 3, "e", -0.018089069918),
        (5, 16, 1, inf, "e", 0.1013825511),
        (5, 16, 2, 2, "e", -0.1566141641),
        (5, 16, 2, 3, "e", -0.080589069918),
        (5, 16, 2, inf, "e", 0.0126694050),
        (6, 4, 1, 2, "e", -0.14070180426292878),
        (6, 4, 1, 3, "e", -0.076275290026),
        (6, 4, 1, inf, "e", 0.0765195523),
        (7, 4, 1, 2, "e", -0.1333354231111923),
        (7, 4, 1, 3, "e", -0.07016025200),
        (7, 4, 1, inf, "e", 0.09183638914294),
        (7, 4, 1, 2, 2.0, -0.19236235),
        (7, 4, 1, 3, 10, -0.03047021),
        (7, 4, 1, inf, 5, 0.057061156),
    ],
)
def test_ksg_te(rng_int, k, ksg_id, minkowski_p, base, expected):
    """Test the KSG transfer entropy estimator."""
    data_source, data_dest = generate_autoregressive_series(rng_int, 0.5, 0.6, 0.4)
    est = KSGTEEstimator(
        data_source,
        data_dest,
        k=k,
        ksg_id=ksg_id,
        minkowski_p=minkowski_p,
        noise_level=0,  # for reproducibility
        base=base,
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,prop_time,step_size,src_hist_len,dest_hist_len,expected",
    [
        (5, 1, 1, 1, 1, -0.005709180172),
        (6, 1, 1, 1, 1, 0.008991300793),
        (7, 1, 1, 1, 1, 0.008242300136),
        (5, 0, 1, 1, 1, 0.09269348225),
        (5, 1, 2, 1, 1, 0.01511405958),
        (5, 1, 3, 1, 1, 0.00351323823),
        (5, 1, 1, 2, 1, 0.00986613205),
        (5, 1, 1, 1, 2, 0.01485397749),
        (5, 1, 1, 2, 2, 0.029176944519),
    ],
)
def test_ksg_te_slicing(
    rng_int,
    prop_time,
    step_size,
    src_hist_len,
    dest_hist_len,
    expected,
):
    """Test the KSG transfer entropy estimator with slicing."""
    data_source, data_dest = generate_autoregressive_series(rng_int, 0.5, 0.6, 0.4)
    est = KSGTEEstimator(
        data_source,
        data_dest,
        prop_time=prop_time,
        step_size=step_size,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        noise_level=0,  # for reproducibility
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,k,ksg_id,minkowski_p,base,expected",
    [
        (5, 4, 1, 2, "e", -0.14824377373692343),
        (5, 4, 2, 2, "e", -0.39824377373692343),
        (5, 4, 1, 3, "e", -0.105711677),
        (5, 4, 1, inf, "e", 0.089420174),
        (5, 16, 1, 2, "e", -0.0629498022),
        (5, 16, 1, 3, "e", -0.017337609),
        (5, 16, 1, inf, "e", 0.082505768),
        (6, 4, 1, 2, "e", -0.1120271189),
        (6, 4, 1, 3, "e", -0.066107282),
        (6, 4, 1, inf, "e", 0.117280006),
        (7, 4, 1, 2, "e", -0.17448870909),
        (7, 4, 1, 3, "e", -0.126082959),
        (7, 4, 1, inf, "e", 0.069872334),
        (7, 4, 1, 2, 2.0, -0.25173399),
        (7, 4, 1, 3, 10, -0.054757133),
        (7, 4, 1, inf, 5.0, 0.043414122),
    ],
)
def test_ksg_cte(rng_int, k, ksg_id, minkowski_p, base, expected):
    """Test the conditional KSG transfer entropy estimator."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_cond,
        k=k,
        ksg_id=ksg_id,
        minkowski_p=minkowski_p,
        noise_level=0,  # for reproducibility
        base=base,
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,step_size,src_hist_len,dest_hist_len,cond_hist_len,ksg_id,expected",
    [
        (5, 1, 1, 1, 1, 1, 0.089420174),
        (5, 1, 1, 1, 1, 2, -0.2817847788),
        (6, 1, 1, 1, 1, 1, 0.1172800069),
        (6, 1, 1, 1, 1, 2, -0.2554896724),
        (7, 1, 1, 1, 1, 1, 0.0698723348),
        (7, 1, 1, 1, 1, 2, -0.2981837603),
        (5, 1, 1, 1, 1, 1, 0.089420174),
        (5, 2, 1, 1, 1, 1, 0.052874143),
        (5, 3, 1, 1, 1, 1, 0.064791017),
        (5, 1, 2, 1, 1, 1, 0.085618601),
        (5, 1, 1, 2, 1, 1, 0.089773402),
        (5, 1, 1, 1, 2, 1, 0.077874619),
        (5, 1, 2, 2, 1, 1, 0.065095054),
        (5, 1, 2, 2, 2, 1, 0.061523405),
    ],
)
def test_ksg_cte_slicing(
    rng_int,
    step_size,
    src_hist_len,
    dest_hist_len,
    cond_hist_len,
    ksg_id,
    expected,
):
    """Test the conditional KSG transfer entropy estimator with slicing."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_cond,
        step_size=step_size,
        src_hist_len=src_hist_len,
        dest_hist_len=dest_hist_len,
        cond_hist_len=cond_hist_len,
        ksg_id=ksg_id,
        noise_level=0,  # for reproducibility
    )
    assert isinstance(est.result(), float)
    assert est.result() == pytest.approx(expected)
    assert isinstance(est.local_vals(), ndarray)


@pytest.mark.parametrize(
    "rng_int,method,p_te,p_cte",
    [
        (1, "permutation_test", 0.0, 0.0),
        (1, "bootstrap", 0.0, 0.0),
        (2, "permutation_test", 0.0, 0.0),
        (2, "bootstrap", 0.0, 0.0),
        (3, "permutation_test", 0.0, 0.0),
        (4, "permutation_test", 0.0, 0.0),
    ],
)
@pytest.mark.parametrize("ksg_id", [1, 2])
def test_ksg_te_statistical_test(rng_int, method, p_te, p_cte, ksg_id):
    """Test the KSG TE for p-values. Fix rng."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est_te_xy = KSGTEEstimator(
        data_source,
        data_dest,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base=2,
        seed=8,
        ksg_id=ksg_id,
    )
    est_cte_xy = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_cond,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base=2,
        seed=8,
        ksg_id=ksg_id,
    )
    test = est_te_xy.statistical_test(method=method, n_tests=50)
    assert test.p_value == pytest.approx(p_te)
    test = est_cte_xy.statistical_test(method=method, n_tests=50)
    assert test.p_value == pytest.approx(p_cte)


@pytest.mark.parametrize(
    "rng_int,method,eff_te_1,eff_te_2,eff_cte_1,eff_cte_2",
    [
        (1, "permutation_test", 0.109132408, 0.102263973, 0.0, 0.0),
        (1, "bootstrap", 0.114234819, 0.108487012, 0.0, 0.0),
        (2, "permutation_test", 0.042919601, 0.044615241, 0.0, 0.0),
        (2, "bootstrap", 0.053020353, 0.054544085, 0.0, 0.0),
        (3, "permutation_test", 0.071168534, 0.067808382, 0.0, 0.0),
        (4, "permutation_test", 0.096478042, 0.089971667, 0.0, 0.0),
    ],
)
def test_ksg_te_effective_val(
    rng_int, method, eff_te_1, eff_te_2, eff_cte_1, eff_cte_2
):
    """Test the KSG transfer entropy for effective values. Fix rng."""
    data_source, data_dest, data_cond = generate_autoregressive_series_condition(
        rng_int, alpha=(0.5, 0.1), beta=0.6, gamma=(0.4, 0.2)
    )
    est_te_xy_1 = KSGTEEstimator(
        data_source,
        data_dest,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base="e",
        seed=8,
        ksg_id=1,
    )
    est_cte_xy_1 = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_source,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base="e",
        seed=8,
        ksg_id=1,
    )
    assert est_te_xy_1.effective_val(method=method) == pytest.approx(eff_te_1)
    assert est_cte_xy_1.effective_val(method=method) == pytest.approx(eff_cte_1)

    est_te_xy_2 = KSGTEEstimator(
        data_source,
        data_dest,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base="e",
        seed=8,
        ksg_id=2,
    )
    est_cte_xy_2 = KSGCTEEstimator(
        data_source,
        data_dest,
        cond=data_source,
        k=4,
        minkowski_p=inf,
        noise_level=0,
        base="e",
        seed=8,
        ksg_id=2,
    )
    assert est_te_xy_2.effective_val(method=method) == pytest.approx(eff_te_2)
    assert est_cte_xy_2.effective_val(method=method) == pytest.approx(eff_cte_2)


def test_ksg_te_variants():
    """Test KSG TE variants."""
    np.random.seed(42)
    x = np.random.rand(100)
    y = np.roll(x, 1) + np.random.normal(0, 0.1, 100)

    te1 = transfer_entropy(x, y, approach="ksg", k=4, ksg_id=1)
    te2 = transfer_entropy(x, y, approach="ksg", k=4, ksg_id=2)

    assert te1 > 0
    assert te2 > 0
    assert te1 != te2


def test_ksg_invalid_id():
    """Test that invalid ksg_id raises ValueError."""
    x = np.random.rand(10)
    y = np.random.rand(10)

    # TE
    with pytest.raises(ValueError, match="ksg_id must be 1 or 2"):
        KSGTEEstimator(x, y, ksg_id=-1)

    # CTE
    with pytest.raises(ValueError, match="ksg_id must be 1 or 2"):
        KSGCTEEstimator(x, y, cond=x, ksg_id=1.5)



"""Ordinal / Permutation utility functions."""

from math import factorial
from typing import Generator

from numpy import (
    ndarray,
    argsort,
    apply_along_axis,
    zeros,
    sum as np_sum,
    array,
    uint64,
    iinfo,
    unique,
    column_stack,
)
from numpy.lib.stride_tricks import as_strided


def permutation_to_integer(perm: ndarray, dtype: type = uint64) -> int:
    """
    Convert a permutation pattern to a unique integer.
    The Lehmer code is used to convert the permutation to an integer.

    Parameters
    ----------
    perm : ndarray
        A permutation pattern.
    dtype : type, optional
        The data type to use for the output. Default is uint64.
        Using this, the maximal number will be set and the smallest possible
        dtype will be used.

    Returns
    -------
    int : int, uint8, uint16, uint32, uint64
        A unique integer representing the permutation pattern.

    Examples
    --------
    >>> permutation_to_integer(array([0, 1]))
    0
    >>> permutation_to_integer(array([1, 0]))
    1
    >>> permutation_to_integer(array([0, 1, 2]))
    0
    >>> permutation_to_integer(array([2, 1, 0]))
    5

    Notes
    -----
    This approach has at least been known since
    1888 :cite:p:`laisantNumerationFactorielleApplication1888`.
    It is named after Derrick Henry Lehmer :cite:p:`Lehmer1960TeachingCT`.

    Raises
    ------
    ValueError
        If the embedding_dim is too large to convert to an uint64 (maximal 20).
    """
    n = len(perm)
    if n > 20:
        raise ValueError(
            "For embedding dimensions larger than 20, "
            "the integer will be too large for uint64."
        )
    factoradic = zeros(n, dtype=dtype)
    for i in range(n):
        factoradic[i] = np_sum(perm[i] > perm[i + 1 :], dtype=dtype)
    integer = np_sum(
        factoradic * array([factorial(n - 1 - i) for i in range(n)]),
        dtype=dtype,
    )
    return integer


def symbolize_series(
    series: ndarray, embedding_dim: int, step_size: int = 1, to_int=False, stable=False
) -> ndarray:
    """
    Convert a time series into a sequence of symbols (permutation patterns).

    Parameters
    ----------
    series : ndarray, shape (n,)
        A numpy array of data points.
    embedding_dim : int
        The size of the permutation patterns.
    step_size : int
        The step size for the sliding windows. Takes every `step_size`-th element.
    to_int : bool, optional
        Whether to convert the permutation patterns to integers. Default is False.
        This
    stable : bool, optional
        If True, when sorting the data, the embedding_dim of equal elements is preserved.
        This can be useful for reproducibility and testing, but might be slower.

    Returns
    -------
    patterns : ndarray, shape (n - (embedding_dim - 1) * step_size, embedding_dim)
        A list of tuples representing the symbolized series.

    Examples
    --------
    >>> series = np.array([1, 2, 3, 2, 1])
    >>> symbolize_series(series, 2, 1)
    array([[0, 1],
           [0, 1],
           [1, 0],
           [1, 0]])

    Raises
    ------
    ValueError
        If the embedding_dim is less than 1.
    ValueError
        If the step_size is less than 1.
    """
    if embedding_dim < 1:
        raise ValueError("The embedding_dim must be a positive integer.")
    if step_size < 1:
        raise ValueError("The step_size must be a positive integer.")
    # Create a view of the series with the given embedding_dim and step size
    shape = (series.size - (embedding_dim - 1) * step_size, embedding_dim)
    strides = (series.strides[0], series.strides[0] * step_size)
    # Extract subsequences
    subsequences = as_strided(series, shape=shape, strides=strides)
    # Get the permutation patterns
    if stable:
        patterns = apply_along_axis(argsort, 1, subsequences, kind="stable")
    else:
        patterns = apply_along_axis(argsort, 1, subsequences)

    # If Lehmer code is requested, convert the permutation to an integer
    if to_int:
        # Determine necessary dtype for the maximal number
        dtypes = ["uint8", "uint16", "uint32"]
        dtype = uint64
        for d in dtypes:  # try small to large
            if factorial(embedding_dim) < iinfo(d).max:
                dtype = d
                break
        # Convert the permutation patterns to integers
        patterns = apply_along_axis(
            lambda x: permutation_to_integer(x, dtype=dtype), 1, patterns
        )

    return patterns


def reduce_joint_space(data: ndarray | tuple[ndarray] | Generator) -> ndarray:
    """Reduce the data to the joint space.

    Reduce features: Assigns each unique feature vector to a unique integer.
    This is equivalent to the unique indices of the unique rows.
    If `data` is 1D, returns the data as is.

    Parameters
    ----------
    data : ndarray, shape (n_samples,) or (n_samples, n_features), or tuple of arrays.
        The data to reduce.

    Returns
    -------
    ndarray, shape (n_samples,)
        The data in the joint space.

    Examples
    --------
    >>> from numpy import array
    >>> data = array([[1, 2], [2, 3], [1, 2], [2, 3], [3, 4]])
    >>> data.shape
    (5, 2)
    >>> reduce_joint_space(data)
    array([0, 1, 0, 1, 2])
    >>> reduce_joint_space(array([4, 5, 4, 5, 6]))
    array([4, 5, 4, 5, 6])
    >>> reduce_joint_space(array([[True, True, False], [False, True, False]]))
    array([0, 1])
    >>> reduce_joint_space(array([[3, 3], [2, 2], [1, 1]]))
    array([2, 1, 0])
    >>> reduce_joint_space((array([3, 2, 1]), array([3, 2, 1])))
    array([2, 1, 0])

    Raises
    ------
    ValueError
        If the data array is not 1D or 2D.
    ValueError
        If the data is a list.

    Notes
    -----
    The order of the unique values is not guaranteed, only the unique indices.
    This is because numpy sorts the values internally.
    """
    if isinstance(data, list):
        raise ValueError("Data array must be a numpy array or tuple of arrays.")
    if isinstance(data, Generator):
        data = tuple(data)
    if isinstance(data, tuple):
        data = column_stack(data)
    if data.ndim == 2:
        # Reduce the data to the joint space using the numpy unique function
        return unique(data, axis=0, return_inverse=True)[1]  # use inverse indices
    elif data.ndim == 1:
        return data
    raise ValueError("Data array must be 1D or 2D.")



"""Module for the discrete mutual information estimator."""

from abc import ABC

from numpy import ndarray

from ..base import (
    MutualInformationEstimator,
    ConditionalMutualInformationEstimator,
)
from ..mixins import DiscreteMIMixin
from ..utils.discrete_interaction_information import (
    mutual_information_global,
    mutual_information_local,
    conditional_mutual_information_global,
    conditional_mutual_information_local,
)
from ..utils.ordinal import reduce_joint_space
from ... import Config
from ...utils.types import LogBaseType


class BaseDiscreteMIEstimator(DiscreteMIMixin, ABC):
    """Base class for discrete mutual information estimators.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the (conditional) mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
        For conditional mutual information, only two data arrays are allowed.
    cond : array-like or tuple of array-like, optional
        The conditional data used to estimate the conditional mutual information.
        Multiple RVs can be passed as tuples and will be treated as joint distributions.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
        Not compatible with the ``cond`` parameter / conditional MI.
    """

    def __init__(
        self,
        *data,
        cond=None,
        offset: int = 0,
        base: LogBaseType = Config.get("base"),
        **kwargs,
    ):
        """Initialize the BaseDiscreteMIEstimator.

        Parameters
        ----------
        *data : array-like, shape (n_samples,)
            The data used to estimate the (conditional) mutual information.
            You can pass an arbitrary number of data arrays as positional arguments.
            For conditional mutual information, only two data arrays are allowed.
        cond : array-like or tuple of array-like, optional
            The conditional data used to estimate the conditional mutual information.
            Multiple RVs can be passed as tuples and will be treated as joint
            distributions.
        offset : int, optional
            Number of positions to shift the X and Y data arrays relative to each other.
            Delay/lag/shift between the variables. Default is no shift.
            Not compatible with the ``cond`` parameter / conditional MI.
        """
        self.data: tuple[ndarray] = None

        if cond is None:
            super().__init__(*data, offset=offset, normalize=False, base=base, **kwargs)
        else:
            if isinstance(cond, (tuple, ndarray)):
                cond = reduce_joint_space(cond)
            super().__init__(
                *data, cond=cond, offset=offset, normalize=False, base=base, **kwargs
            )
        self._check_data_mi()


class DiscreteMIEstimator(BaseDiscreteMIEstimator, MutualInformationEstimator):
    """Estimator for the discrete mutual information.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    offset : int, optional
        Number of positions to shift the data arrays relative to each other.
        Delay/lag/shift between the variables. Default is no shift.
    """

    def _calculate(self):
        """Calculate the mutual information of the data.

        Returns
        -------
        float
            The calculated mutual information.
        """
        return mutual_information_global(*self.data, log_func=self._log_base)

    def _extract_local_values(self) -> ndarray[float]:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of mi.
        """
        return mutual_information_local(*self.data, log_func=self._log_base)


class DiscreteCMIEstimator(
    BaseDiscreteMIEstimator, ConditionalMutualInformationEstimator
):
    """Estimator for the discrete conditional mutual information.

    Attributes
    ----------
    *data : array-like, shape (n_samples,)
        The data used to estimate the conditional mutual information.
        You can pass an arbitrary number of data arrays as positional arguments.
    cond : array-like
        The conditional data used to estimate the conditional mutual information.
    """

    def _calculate(self):
        """Calculate the conditional mutual information of the data.

        Returns
        -------
        float
            The calculated conditional mutual information.
        """
        return conditional_mutual_information_global(
            *self.data, cond=self.cond, log_func=self._log_base
        )

    def _extract_local_values(self) -> ndarray:
        """Separately calculate the local values.

        Returns
        -------
        ndarray[float]
            The calculated local values of cmi.
        """
        return conditional_mutual_information_local(
            *self.data, cond=self.cond, log_func=self._log_base
        )
